package dev.pfaff.altargui.types;

import static java.util.FormatProcessor.FMT;

/**
 * A type representing an RGBA color with 8-bit channels.
 * <p>
 * Until value classes are available for use, avoid storing instances except in {@code static final} or other
 * long-lived fields.
 */
// TODO: valhalla value class
public record Color(int rgba) {
	public static Color WHITE = new Color(0xffffffff);
	public static Color TRANSPARENT = new Color(0x00000000);

	public static Color fromRgba(int rgba) {
		return new Color(rgba);
	}

	private static String makeExtraBitsMessage(int value, int upr, int lwr) {
		return new StringBuilder()
				.append("extra bits in [")
				.append(upr)
				.append(", ")
				.append(lwr)
				.append("]: ")
				.repeat('0', Integer.numberOfLeadingZeros(value))
				.append(Integer.toBinaryString(value))
				.toString();
	}

	public static Color fromRgba(int r, int g, int b, int a) {
		assert (r & 0xffffff_00) == 0 : makeExtraBitsMessage(r, 31, 8);
		assert (g & 0xffffff_00) == 0 : makeExtraBitsMessage(g, 31, 8);
		assert (b & 0xffffff_00) == 0 : makeExtraBitsMessage(b, 31, 8);
		assert (a & 0xffffff_00) == 0 : makeExtraBitsMessage(a, 31, 8);
		return fromRgba((r << 24) | (g << 16) | (b << 8) | a);
	}

	public static Color fromRgb(int rgb) {
		assert (rgb & 0xff_000000) == 0 : makeExtraBitsMessage(rgb, 31, 24);
		return fromRgba(rgb << 8 | 0xff);
	}

	public static Color fromRgb(int r, int g, int b) {
		return fromRgba(r, g, b, 0xff);
	}

	public static Color fromArgb(int argb) {
		return fromRgba(Integer.rotateLeft(argb, 8));
	}

	/**
	 * Returns the colorRgba in ARGB format.
	 */
	public int argb() {
		return Integer.rotateRight(rgba, 8);
	}

	public int abgr() {
		return Integer.reverseBytes(rgba());
	}

	public int red() {
		return (rgba & 0xff000000) >>> 24;
	}

	public int green() {
		return (rgba & 0xff0000) >>> 16;
	}

	public int blue() {
		return (rgba & 0xff00) >>> 8;
	}

	public int alpha() {
		return rgba & 0xff;
	}

	public int rgb() {
		return rgba >>> 8;
	}

	public Color withAlpha(int a) {
		assert (a & 0xffffff_00) == 0 : makeExtraBitsMessage(a, 31, 8);
		return fromRgba((rgba & 0xffffff00) | a);
	}

	public Color premultiply() {
		var alpha = alpha();
		return Color.fromRgba(red() * alpha / 255, green() * alpha / 255, blue() * alpha / 255, alpha);
	}

	public Color unpremultiply() {
		var alpha = alpha();
		if (alpha == 0) return this;
		return Color.fromRgba(red() * 255 / alpha, green() * 255 / alpha, blue() * 255 / alpha, alpha);
	}

	@Override
	public String toString() {
		return FMT."#%08x\{rgba}";
	}
}
