package dev.pfaff.altargui.widgets;

import dev.pfaff.altargui.StateContext;
import dev.pfaff.altargui.Widget;
import dev.pfaff.altargui.WidgetInstance;
import dev.pfaff.altargui.WidgetInstantiationArgs;
import dev.pfaff.altargui.widgets.SignalBuilderWidget.SignalManager;
import dev.pfaff.altargui.widgets.state.StatefulWidgetInstance;
import dev.pfaff.altargui.util.listenable.ListenerDisposer;
import dev.pfaff.altargui.util.listenable.Value;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.invoke.MethodHandles;
import java.util.function.Function;

/**
 * A widget that replaces its child using the provided {@code builder} whenever notified by signal.
 */
public record ValueBuilderWidget(Value<@NotNull Widget> signal) implements Widget {
	public static <T> ValueBuilderWidget builder(Value<T> listenable, Function<T, @NotNull Widget> builder) {
		return new ValueBuilderWidget(listenable.map(builder));
	}

	@Override
	public @NotNull WidgetInstance<ValueBuilderWidget> instantiate(WidgetInstantiationArgs args) {
		return new State(args);
	}

	private static final class State extends StatefulWidgetInstance<ValueBuilderWidget> {
		private static final SignalManager<State, Value<@NotNull Widget>> SIGNAL_MANAGER = new SignalManager<>(
				MethodHandles.lookup(),
				"signal", (Class<Value<@NotNull Widget>>) (Class) Value.class,
				"listener",
				(state, signal) -> signal.listenAndInvoke(child -> {
					//noinspection ConstantValue
					if (child == null) {
						// using this.signal here to avoid another capture.
						throw new NullPointerException(STR."\{ValueBuilderWidget.class.getName()} signal emitted a null value: \{state.signal}");
					}
					state.child = child;
					state.markNeedsRebuild();
				})
		);

		private Value<@NotNull Widget> signal;
		private Widget child;

		private @Nullable ListenerDisposer listener;

		private State(WidgetInstantiationArgs args) {
			super(args);
		}

		@Override
		protected @NotNull Widget build() {
			return this.child;
		}

		@Override
		public void configure(StateContext stateContext, ValueBuilderWidget widget, boolean init) {
			SIGNAL_MANAGER.configure(this, widget.signal);
		}

		@Override
		protected void onMount() {
			super.onMount();

			SIGNAL_MANAGER.onMount(this);
		}

		@Override
		protected void onUnmount() {
			SIGNAL_MANAGER.onUnmount(this);
			super.onUnmount();
		}
	}
}
